﻿//-----------------------------------------------------------------------------------------------------------
//
// THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
// KIND, WHETHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
// PURPOSE.
//
//------------------------------------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;

namespace LinqToHbaseProvider
{
    class HBaseQueryContext
    {
        private HBaseQueryContext() { }

        // Executes the expression tree that is passed to it.
        internal static object Execute(Expression expression, bool IsEnumerable)
        {
            // The expression must represent a query over the data source.
            if (!IsQueryOverDataSource(expression))
                throw new InvalidProgramException("No query over the data source was specified.");

            // Find the call to Where() and get the lambda expression predicate.
            InnermostWhereFinder whereFinder = new InnermostWhereFinder();
            MethodCallExpression whereExpression = whereFinder.GetInnermostWhere(expression);
            LambdaExpression lambdaExpression = (LambdaExpression)((UnaryExpression)(whereExpression.Arguments[1])).Operand;

            // Send the lambda expression through the partial evaluator.
            lambdaExpression = (LambdaExpression)Evaluator.PartialEval(lambdaExpression);


            // Get the column names 
            ColumnFinder cf = new ColumnFinder(lambdaExpression.Body);
            List<string> columns = cf.Columns;
            if (columns.Count == 0)
                throw new InvalidQueryException("You must specify at least one column in your query.");

            AccessRecord[] records;
            if (expression.ToString().Contains("SelectMany") && expression.ToString().Contains("HBase.frequency"))
            {
                List<AccessRecord> aList = new List<AccessRecord>();
                aList.Add(new AccessRecord(columns.First<string>(), (cf.IsIP ? "ip" : "page")));
                records = aList.ToArray<AccessRecord>();
            }
            else
            {
                // Call HBase and get the results.
                records = HBaseHelper.GetColumnsFromHBase(columns.First<string>(), cf.IsIP, cf.IsPage);
            }

            // Copy the IEnumerable places to an IQueryable.
            IQueryable<AccessRecord> queryableRecords = records.AsQueryable<AccessRecord>();

            // Copy the expression tree that was passed in, changing only the first
            // argument of the innermost MethodCallExpression.
            ExpressionTreeModifier treeCopier = new ExpressionTreeModifier(queryableRecords);
            Expression newExpressionTree = treeCopier.CopyAndModify(expression);

            // This step creates an IQueryable that executes by replacing Queryable methods with Enumerable methods.
            if (IsEnumerable)
                return queryableRecords.Provider.CreateQuery(newExpressionTree);
            else
                return queryableRecords.Provider.Execute(newExpressionTree);
        }

        private static bool IsQueryOverDataSource(Expression expression)
        {
            // If expression represents an unqueried IQueryable data source instance,
            // expression is of type ConstantExpression, not MethodCallExpression.
            return (expression is MethodCallExpression);
        }
    }
}
